/* --------------------------------------------------------------------------*
*                  OpenSim:  Contact Force Example                           *
* -------------------------------------------------------------------------- *
*/

#include <OpenSim/OpenSim.h>
#include <Simbody.h>
#include "Simbody.h"
#include "C:\OpenSim 3.3\sdk\include\OpenSim\Simulation\Control\Controller.h"


using namespace OpenSim;
using namespace SimTK;

class PDController2 : public Controller
{
	OpenSim_DECLARE_CONCRETE_OBJECT(PDController2, Controller);

private:
	double kp;
	double kv;

public:

	PDController2::PDController2() : Controller() {
		setKp(100);
		setKv(20);
	}

	void PDController2::setKp(double kp_value) {
		kp = kp_value;
	}

	void PDController2::setKv(double kv_value) {
		kv = kv_value;
	}

	Vector PDController2::VelocityVector(const State& s) const {

		Vector c;

		_model->getMultibodySystem().realize(s, SimTK::Stage::Velocity);
		_model->getMatterSubsystem().calcResidualForceIgnoringConstraints(s, SimTK::Vector(0), SimTK::Vector_<SimTK::SpatialVec>(0), SimTK::Vector(0), c);

		return c;
	}

	Vector PDController2::C21(const State& s) const {
		Vector c21(2, Real(0));
		Vector c31 = VelocityVector(s);

		for (unsigned int iJ = 0; iJ < 2; iJ++) {
			c21[iJ] = c31[iJ];
		};

		return c21;
	}

	Vector PDController2::GravityVector(const State& s) const {
		Vector g;

		_model->getMatterSubsystem().multiplyBySystemJacobianTranspose(s, _model->getGravityForce().getBodyForces(s), g);

		return g;
	}

	Vector PDController2::G21(const State& s) const {
		Vector g21(2, Real(0));
		Vector g31 = GravityVector(s);

		for (unsigned int iJ = 0; iJ < 2; iJ++) {
			g21[iJ] = g31[iJ];
		};

		return g21;
	}

	Matrix PDController2::MassMatrix(const State& s) const {
		int numMobilities = s.getNU();
		Matrix M(numMobilities, numMobilities);

		_model->getMatterSubsystem().calcM(s, M);

		return M;

		//		std::cout << "M: " << M << std::endl;
	}

	Matrix PDController2::Mass22(const State& s) const {
		Matrix Mass33 = MassMatrix(s);
		//	Vector f(7, Real(0));
		Matrix Mass22(2,2);
		//	Vector Mass97col;

		for (unsigned int iT = 0; iT < 2; iT++) {

			for (unsigned int iJ = 0; iJ < 2; iJ++) {
				Mass22[iT][iJ] = Mass33[iT][iJ];
			};
		};
		return Mass22;
	}

	Array<double> PDController2::ForceValue(const State& s) const {
		
		_model->getMultibodySystem().realize(s, SimTK::Stage::Dynamics);
		ForceSet forces = _model->getForceSet();
	//	std::cout << "forceset: " << forces << std::endl;
		OpenSim::Force& contact = forces.get(3);

	//	std::cout << "contactforce: " << contact << std::endl;
		Array<double> F;

	//	Array<std::string> labels = contact.getRecordLabels();
		F = contact.getRecordValues(s);

	//	std::cout << "force labels: " << F << std::endl;
		return F;
	} 


	//_____________________________________________________________________________
	void PDController2::computeControls(const SimTK::State& s, SimTK::Vector &controls) const
	{
		double t = s.getTime();


		// THE GOAL OF THIS EXAMPLE IS JUST TO DEMONSTRATE HOW I CAN OBTAIN VALUES AND POINT OF ACTION OF //
		// CONTACT FORCES. OTHER THINGS CAN BE IGNORED. //

		Array<double> force = ForceValue(s);

		std::cout << "This is the force of interest: " << force << std::endl;

		const Actuator& j1_act = _model->getActuators().get("joint_1_actuator");
		const Actuator& j2_act = _model->getActuators().get("joint_2_actuator");

		double control_1 = 0.5;
		double control_2 = 0.5;

		Vec7 control(control_1, control_2);

		Vector control1(1, control[0]);
		Vector control2(1, control[1]);

		j1_act.addInControls(control1, controls);
		j2_act.addInControls(control2, controls);

	}
};

int main()
{
	try {
		// Create an OpenSim model and set its name
//		Model osimModel("Two_Link_Arm.osim");
		Model osimModel("TwoLinkArmModel.osim");
		osimModel.setUseVisualizer(true);

		double kp = 100.0;
		double kv = 20.0;

		PDController2 *controller = new PDController2();
		controller->setName("my_controller");
		controller->setActuators(osimModel.updActuators());
		osimModel.addController(controller);

		State& s = osimModel.initSystem();

		CoordinateSet& modelSet = osimModel.updCoordinateSet();

		const Coordinate& joint1 = osimModel.getCoordinateSet().get("joint_1");
		const Coordinate& joint2 = osimModel.getCoordinateSet().get("joint_2");

		joint1.setValue(s, 90*Pi/180);
		joint1.setSpeedValue(s, 0);
		joint2.setValue(s, 60*Pi/180);
		joint2.setSpeedValue(s, 0);

		std::cout << "System Initialized" << std::endl;

		std::cout << "getMultibodySystem Done" << std::endl;

		SimTK::RungeKuttaMersonIntegrator integrator(osimModel.getMultibodySystem());
		integrator.setAccuracy(1.0e-3);
		Manager manager(osimModel, integrator);

		// Define the initial and final simulation times
		double initialTime = 0.0;
		double finalTime = 1.0;

		osimModel.updMatterSubsystem().setShowDefaultGeometry(true);
		Visualizer& viz = osimModel.updVisualizer().updSimbodyVisualizer();

		// Integrate from initial time to final time
		manager.setInitialTime(initialTime);
		manager.setFinalTime(finalTime);
		std::cout << "\n\nIntegrating from " << initialTime << " to " << finalTime << std::endl;
		manager.integrate(s);

	}
	catch (OpenSim::Exception ex)
	{
		std::cout << ex.getMessage() << std::endl;
		return 1;
	}
	catch (std::exception ex)
	{
		std::cout << ex.what() << std::endl;
		return 1;
	}
	catch (...)
	{
		std::cout << "UNRECOGNIZED EXCEPTION" << std::endl;
		return 1;
	}

	std::cout << "OpenSim example completed successfully.\n";
	std::cin.get();
	return 0;
}



